/*
 * rrlms.c - Road Runner login program
 *
 * 6-16-98
 *
 * Larkin Lowrey - llowrey@NuclearWinter.com
 * http://Larkin.NuclearWinter.com/rros2/
 *
 * This is freeware.
 *
 * See http://Larkin.NuclearWinter.com/rros2/Digital.pdf for information
 * on the Digital authentication protocol
 */

/*
 * mods by George Coulouris <glc5@cornell.edu>
 * http://www.tc.cornell.edu/~glc5/
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <pwd.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include "global.h"
#include "md5.h"

void sighup_handler(int s)
{
   fprintf(stderr,"caught sighup, killing myself.\n");
   remove("/var/run/rrlms.pid");
   exit(1);
}

// This implements an experimental Digital authentication RR login

// rrlms username password <ip> <port>

char in_packet[1024], out_packet[1024];
char hashed[16], hashsrc[1024];
char OS_Name[20], OS_Version[20], szAuthAddr[64];
unsigned int AuthAddr, ProtocolAddr;
unsigned short AuthPort, ProtocolPort, VersionID;

void sighandler(int val)
{
	printf("\nDNS lookup timed out.\n");
	printf("Make sure that you have the DNS server address set correctly.\n\n");
	exit(-1);
}

int PrintRawPacket(unsigned char *buf, int len)
{
	int loop;

	printf("Raw Packet Data:\n");
	for (loop=0;loop<len;loop++)
	{
		if (len%16==0)
			printf("\n");
		printf("%02X",buf[loop]);
	}
	printf("\n");
	fflush(stdout);
	return 0;
}

int Lookup(char *name, unsigned int *ip)
{
    struct hostent *he;

    if (!ip)
        return -1;

	alarm(30);
	he=gethostbyname(name);
	alarm(0);
    if (!he)
        return -2;
	
    *ip=*(int *)he->h_addr_list[0];

    return 1;
}

int NewPacket(short type)
{
	*(short *)out_packet=htons(type);
	*(short *)(out_packet+2)=htons(8);
	*(short *)(out_packet+4)=0;
	*(short *)(out_packet+6)=0;

	return 0;
}

int NewParam(short type, char *buffer, short len)
{
	short packetlen;

	packetlen=ntohs(*(short *)(out_packet+2));

	*(short *)(out_packet+packetlen)=htons(type);
	memcpy(out_packet+packetlen+4,buffer,len);
	*(short *)(out_packet+packetlen+2)=htons(len+4);
	*(short *)(out_packet+2)=htons(packetlen+len+4);

	return 0;
}

int GetData(short type, unsigned char *buffer)
{
	unsigned char *c, *max;

	c=in_packet+8;
	max=c+ntohs(*(short *)(in_packet+2))-8;
	while (*(short *)c!=htons(type) && c<max)
		c+=ntohs(*(short *)(c+2));
	
	if (c>=max)
		return -1;

	memcpy(buffer,c+4,ntohs(*(short *)(c+2))-4);
	return 0;
}

int GetString(short type, unsigned char *buffer)
{
	unsigned char *c, *max;

	c=in_packet+8;
	max=c+ntohs(*(short *)(in_packet+2))-8;
	while (*(short *)c!=htons(type) && c<max)
		c+=ntohs(*(short *)(c+2));
	
	if (c>=max)
		return -1;

	memcpy(buffer,c+4,ntohs(*(short *)(c+2))-4);
	buffer[ntohs(*(short *)(c+2))-4]=0;
	return 0;
}

int PrintPacket(char *packet)
{
	short packetlen, len;
	unsigned char *c;

	len=ntohs(*(short *)(packet+2));
	c=packet;
	while (len--)
	{
		printf("%02X",*c);
		c++;
	}
	printf("\n\n");

	packetlen=ntohs(*(short *)(packet+2));

	printf("Packet Type: %d\n",ntohs(*(short *)packet));
	fflush(stdout);
	printf("Packet Length: %d\n",packetlen);

	c=(unsigned char *)(packet+8);
	packetlen-=8;

	while (packetlen)
	{
		printf("\tType: %d\n",ntohs(*(short *)c));
		c+=2;packetlen-=2;
		printf("\tLength: %d\n",ntohs(*(short *)c));
		len=ntohs(*(short *)c)-4;
		c+=2;packetlen-=2;
		printf("\t\t0x");
		while (len)
		{
			printf("%02X",*(unsigned char *)c);
			c++;packetlen--;len--;
		}
		printf("\n");
	}
	printf("\n");
	fflush(stdout);
	return 0;
}

int Hash(unsigned char *target, unsigned char *source, int len)
{
	MD5_CTX context;
	unsigned char *c;
	int len1;

	c=source;
	len1=len;
	printf("Hashing 0x");
	while (len1--)
	{
		printf("%02X",*c);
		c++;
	}
	printf("\n\n");


	memset((char *)&context,0,sizeof(context));
	MD5Init(&context);
	MD5Update(&context,source,len);
	MD5Final(target,&context);
}

int ProtocolNegotiation(char *server)
{
	int sock,rc;
	struct sockaddr_in sin;
	unsigned short num;

	if (Lookup(server,&ProtocolAddr)<0)
	{
		printf("Unable to resolve IP for protocol-server\n");
		return -1;
	}

	if ((sock=socket(AF_INET,SOCK_STREAM,0))==-1)
	{
		printf("Failed to allocate socket\n");
		return -2;
	}

	sin.sin_family=AF_INET;
	sin.sin_port=htons(60000);
	sin.sin_addr.s_addr=ProtocolAddr;

	if (connect(sock,(struct sockaddr *)&sin,sizeof(sin))==-1)
	{
		printf("Unable to connect to server\n");
		close(sock);
		return -3;
	}

	NewPacket(1);
	NewParam(3,(char *)&VersionID,sizeof(VersionID));
	NewParam(4,OS_Name,strlen(OS_Name));
	NewParam(5,OS_Version,strlen(OS_Version));
	num=htons(1);
	NewParam(1,(char *)&num,sizeof(num));

	PrintPacket(out_packet);
    rc=send(sock,out_packet,ntohs(*(short *)(out_packet+2)),0);
	printf("Send RC=%d\n",rc);
	if (rc==-1)
		return -4;

	rc=recv(sock,in_packet,1024,0);
	printf("Recieve RC=%d\n",rc);
	PrintRawPacket(in_packet,rc);
	PrintPacket(in_packet);
	if (rc==-1)
		return -5;

	if (*(short *)in_packet!=htons(2))
	{
		printf("Unexpected packet type\n");
		return -6;
	}

	if (GetData(10,(unsigned char *)&num)<0 || num)
	{
		printf("Invalid response\n");
		return -7;
	}

	GetString(24,szAuthAddr);
	GetData(15,(unsigned char *)&AuthPort);

	printf("Authentication server %s:%d\n",szAuthAddr,ntohs(AuthPort));

	return 0;
}
	
void main(int argc, char **argv)
{
   char *username, *password, *server;
   int serverport;

   short num;
   struct sockaddr_in sin;
   int sock, rc, sinlen, kasock, sequence, now, bignum;
   unsigned int authip, blinding;
   unsigned short authport, HashPWD, kaport;
   unsigned char Nonce[16], *uc;
   struct tm *tm;
   
   pid_t mypid;
   FILE *f;
   
   VersionID=htons(11);
   strcpy(OS_Name,"Windows 95");
   strcpy(OS_Version,"4.0");

   username = (char *) malloc(64);
   password = (char *) malloc(64);
   server   = (char *) malloc(64);
   f=fopen("/etc/rrlms.conf","r");
   
   if (f == NULL)
     {
	fprintf(stderr,"Couldn't open /etc/rrlms.conf\n");
	exit(1);
     }
   else
     {
	fscanf(f,"%s %s %s",server, username, password);
	fprintf(stderr,"server: %s, username: %s, password: %s\n",server, username,password);
	fclose(f);
     }
   
   f=fopen("/var/run/rrlms.pid","r");
   if (f != NULL)
     {
	fprintf(stderr,"/var/run/rrlms.pid exists. another instance running?\n");
	fclose(f);
	exit(1);
     }
      
   mypid=getpid();
   f=fopen("/var/run/rrlms.pid","w");
   fprintf(f,"%d\n",mypid);
   fclose(f);
   
   signal(SIGHUP, sighup_handler);
   
	if (ProtocolNegotiation(server)<0)
	{
		printf("Protocol Negotiation failed. Unable to proceed\n");
		exit(-1);
	}

	printf("lookup %s\n",szAuthAddr);
	if (Lookup(szAuthAddr,&AuthAddr)<0)
	{
		printf("Authentication host, %s, could not be located\n",szAuthAddr);
		exit(-1);
	}

	if ((kasock=socket(AF_INET,SOCK_DGRAM,0))==-1)
	{
		printf("Unable to allocate Keep-Alive socket\n");
		exit(-1);
	}
	
	sin.sin_family=AF_INET;
	sin.sin_port=0;
	sin.sin_addr.s_addr=0;

	bind(kasock,(struct sockaddr *)&sin,sizeof(sin));

	sinlen=sizeof(sin);
	getsockname(kasock,(struct sockaddr *)&sin,&sinlen);
	kaport=sin.sin_port;

	if ((sock=socket(AF_INET,SOCK_STREAM,0))==-1)
	{
		printf("Unable to allocate socket\n");
		exit(-1);
	}

	sin.sin_family=AF_INET;
	sin.sin_port=AuthPort;
	sin.sin_addr.s_addr=AuthAddr;

	uc=(unsigned char *)&sin.sin_addr.s_addr;
	printf("Connecting to: %d.%d.%d.%d:",uc[0],uc[1],uc[2],uc[3]);
	uc=(unsigned char *)&sin.sin_port;
	printf("%d\n",ntohs(*(short *)uc));
    if (connect(sock,(struct sockaddr *)&sin,sizeof(sin))<0)
	{
		printf("Unable to connect to auth server\n");
		exit(-4);
	}

	NewPacket(3);
	NewParam(7,username,strlen(username));
	NewParam(3,(char *)&VersionID,sizeof(VersionID));
	NewParam(4,OS_Name,strlen(OS_Name));
	NewParam(5,OS_Version,strlen(OS_Version));
	num=0;
	NewParam(6,(char *)&num,sizeof(num));
	NewParam(8,(char *)&kaport,sizeof(kaport));

	printf("Sending...\n");
	PrintPacket(out_packet);
    rc=send(sock,out_packet,ntohs(*(short *)(out_packet+2)),0);
	printf("SendRC=%d\n",rc);

	if ((rc=recv(sock,in_packet,1024,0))<8)
	{
		printf("Error recieving from auth server; rc=%d\n",rc);
		exit(-4);
	}
	printf("Recieved...rc=%d\n",rc);
	fflush(stdout);
	PrintRawPacket(in_packet,rc);
	PrintPacket(in_packet);

	if (*(short *)in_packet!=htons(9))
	{
		printf("Packet type not expected\n");
		exit(-5);
	}

	GetData(0x000E,(char *)&HashPWD);
	GetData(0x000C,Nonce);

	blinding=htonl(time(NULL));

	if (HashPWD)
	{
		memcpy(hashsrc,Nonce,16);
		Hash(hashsrc+16,password,strlen(password));
		memcpy(hashsrc+16+16,(char *)&blinding,4);
		hashsrc[16+16+4]=0;
		hashsrc[16+16+4+1]=4;
		memset(hashed,0,16);
		Hash(hashed,hashsrc,16+16+4+2);
	}
	else
	{
		memcpy(hashsrc,Nonce,16);
		strcpy(hashsrc+16,password);
		memcpy(hashsrc+16+strlen(password),(char *)&blinding,4);
		hashsrc[16+strlen(password)+4]=0;
		hashsrc[16+strlen(password)+4+1]=4;
		Hash(hashed,hashsrc,16+strlen(password)+4+2);
	}

	NewPacket(4);
	NewParam(11,hashed,16);
	NewParam(21,(char *)&blinding,4);

	
	printf("Sending...\n");fflush(stdout);
	PrintPacket(out_packet);
    rc=send(sock,out_packet,ntohs(*(short *)(out_packet+2)),0);
	printf("SendRC=%d\n",rc);

	if ((rc=recv(sock,in_packet,1024,0))<0)
	{
		printf("Error recieving from auth server; rc=%d\n",rc);
		exit(-4);
	}
	printf("Recieved...rc=%d\n",rc);
	fflush(stdout);
	PrintRawPacket(in_packet,rc);
	PrintPacket(in_packet);

	if (*(short *)in_packet!=htons(5))
	{
		printf("unexpected response\n");
		exit(-5);
	}

	rc=time(NULL);
	tm=localtime((const time_t *)&rc);
	printf("Logged in @ %s",asctime(tm));fflush(stdout);

	printf("Keep-alive UDP port %d\n",ntohs(kaport));

	sequence=0;
	while(1)
	{
		sinlen=sizeof(sin);
		rc=recvfrom(kasock,in_packet,1024,0,(struct sockaddr *)&sin,&sinlen);
		now=time(NULL);
		tm=localtime((const time_t *)&now);
		printf("Recieved %d bytes from %s:%d\n @ %s",rc,inet_ntoa(sin.sin_addr),ntohs(sin.sin_port),asctime(tm));
		fflush(stdout);
		PrintRawPacket(in_packet,rc);
		PrintPacket(in_packet);


		if (*(short *)in_packet!=htons(11))
		{
			printf("Unexpected packet\n");
			continue;
		}

		NewPacket(12);
		num=0;
		NewParam(10,(unsigned char *)&num,sizeof(num));

		sequence++;
		bignum=htonl(sequence);
		if (HashPWD)
		{
			memcpy(hashsrc,Nonce,16);
			Hash(hashsrc+16,password,strlen(password));
			memcpy(hashsrc+16+16,(char *)&bignum,4);
			hashsrc[16+16+4]=0;
			hashsrc[16+16+4+1]=12;
			memset(hashed,0,16);
			Hash(hashed,hashsrc,16+16+4+2);
		}
		else
		{
			memcpy(hashsrc,Nonce,16);
			strcpy(hashsrc+16,password);
			memcpy(hashsrc+16+strlen(password),(char *)&bignum,4);
			hashsrc[16+strlen(password)+4]=0;
			hashsrc[16+strlen(password)+4+1]=12;
			Hash(hashed,hashsrc,16+strlen(password)+4+2);
		}
		NewParam(19,hashed,16);
		NewParam(13,(unsigned char *)&bignum,4);

		printf("Sending...\n");fflush(stdout);
		PrintPacket(out_packet);
		sin.sin_port=htons(60003);
   		rc=sendto(kasock,out_packet,ntohs(*(short *)(out_packet+2)),0,(struct sockaddr *)&sin,sizeof(sin));
		printf("SendRC=%d\n",rc);
	}

	exit(0);
}
